from pathlib import Path


def make_path(path: str) -> str:
    """
    Really roundabout way of processing paths for `options.local.remotes_dir`.
    """
    if path == "":
        return ""
    else:
        ppath = Path(path).expanduser()
        if ppath.suffix in {".git"}:
            return str(ppath.with_suffix(""))
        else:
            return str(ppath)


def local_dest_path(
    user_input: str,
    remotes_dir: str,
    host: str,
    path: str,
    ignore_config: bool
) -> str:
    """
    Get the path where the remote repository should be cloned to
    from a variety of inputs.
    
    Parameters
    ----------
    user_input: str
        The destination path given by the user.
    
    remotes_dir: str
        The directory to local clones of remote repositories.
    
    host: str
        The host name of the site storing the remote repository.
    
    path: str
        The path to the remote repository (in the URL).
    
    ignore_config: bool
        Whether to ignore the `options.local.remotes_dir` config.
    
    Returns
    -------
    str
        The destination path to the cloned repository.
    """
    if ignore_config:
        return make_path(user_input)
    else:
        if user_input != "":
            return make_path(user_input)
        else:
            return make_path(Path(remotes_dir) / Path(host) / Path(path))



from __future__ import annotations
import typing as t

from .locators import *
from .scp import *


SCM_WITH_EXPLICIT_SCP: t.Set[str] = {"git"}
"""
A set of version control systems that I currently know that support SCP
locators.
"""


def remote_to_string(remote: UniformResourceLocator, scm: str) -> str:
    """
    Convert a locator to string form.
    
    Parameters
    ----------
    remote: UniformResourceLocator
        The locator identifying the remote repository.
    
    scm: str
        The version control system used to manage the remote repository.
    
    Returns
    -------
    str
        The locator in string form.
    """
    if (
        remote.get_scheme() == "ssh" and
        remote.kwargs.get("explicit_scp") and
        scm in SCM_WITH_EXPLICIT_SCP
    ):
        scp_locator = ScpLocator.from_locator(remote)
        if remote.get_username() == "": # Separate if statements for other SCMs.
            if scm == "git":
                scp_locator.username = "git"
        return str(scp_locator)
    else:
        return str(remote)



import shlex
import shutil
import subprocess
import typing as t

from quickclone.delegation.errors import CommandNotFoundError


__all__ = ["BaseCommand", "Command"]


class BaseCommand(object):
    """
    Base class for representing commands.
    """
    
    def __init__(self, location: str = "", *args: t.Any, **kwargs: t.Any) -> None:
        self.location = location
        self.args = args
        self.kwargs = kwargs
    
    def format_command_list(self) -> t.List[str]:
        """
        Convert this command into a list of command-line arguments.
        
        Returns
        -------
        List[str]
            A list of arguments in the command.
        """
        return []
    
    def format_command_str(self) -> str:
        """
        Convert this command into command-line arguments joined together as a
        single string.
        
        Returns
        -------
        str
            This command's command-line arguments as a single string.
        """
        return " ".join(map(shlex.quote, self.format_command_list()))
        # return shlex.join(self.format_command_list()) # >= 3.8
    
    def run(self) -> t.Union[subprocess.CompletedProcess, subprocess.SubprocessError]:
        """
        Run the command represented by this object using Python's subprocess
        module and return the result from `subprocess.run`.
        
        Returns
        -------
        subprocess.CompletedProcess | subprocess.SubprocessError
            The result of running the command.
        """
        cl = self.format_command_list()
        try:
            process = subprocess.run(cl)
        except subprocess.SubprocessError as se:
            return se
        else:
            return process


class Command(BaseCommand):
    """
    Basic command. Uses `echo` as a dummy command.
    """
    
    COMMAND_NAME: str = "echo" # Dummy command
    
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        location = shutil.which(self.COMMAND_NAME)
        if location is None:
            raise CommandNotFoundError(self.COMMAND_NAME)
        super().__init__(location, *args, **kwargs)
    
    def format_command_list(self) -> t.List[str]:
        kwarg_decomposed = []
        for flag, argument in self.kwargs:
            kwarg_decomposed.extend([flag, argument])
        return [self.location, *self.args, *kwarg_decomposed]



import typing as t

from .common import Command


class MercurialCloneCommand(Command):
    """
    A class representing a `hg clone` command.
    """
    
    COMMAND_NAME: str = "hg"
    
    def __init__(self, remote: str, dest_path: str, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self.remote = remote
        self.dest_path = dest_path
    
    def format_command_list(self) -> t.List[str]:
        cl = super().format_command_list()
        assert len(cl) >= 1
        inject = ["clone", self.remote]
        if self.dest_path != "":
            inject.append(self.dest_path)
        rcl = cl[:1] + inject + cl[1:] 
        return rcl



class CommandNotFoundError(Exception):
    """
    Error for when a command could not be found in the current environment.
    """
    
    def __init__(self, command: str) -> None:
        self.command = command
    
    def __str__(self) -> str:
        return f"'{self.command}' could not be found."


class InvalidVcsError(Exception):
    """
    This error is raised when an invalid version control system is used.
    """
    
    def __init__(self, vcs: str) -> None:
        self.vcs = vcs
    
    def __str__(self) -> str:
        return (
            f"'{self.vcs}' is not a valid version control system or "
            "is not supported by QuickClone yet."
        )



from __future__ import annotations
import re
import typing as t


def extract_groups(
    matches: re.Match,
    names: t.Iterable[str],
    none_str: str = "ignore",
    combine: t.Optional[t.Mapping[str, t.List[str]]] = None
) -> t.Dict[str, t.Optional[str]]:
    """
    Extract named groups from a `re.Match` object to a dictionary.
    These named groups are delimited by `(?P<name>)`.
    
    Parameters
    ----------
    matches: re.Match
        The matches created by `re.search`.
    
    names: Iterable[str]
        A list of named groups found in the regex string.
    
    none_str: str = "ignore"
        If "ignore", strings will be left as they are.
        If "to_none", empty strings will be converted to `None`.
        If "to_str", `None` will be converted to empty strings.
    
    combine: Optional[Mapping[str, List[str]]] = None
        An optional list of named groups that are supposed to be treated as the
        same named group.
        For example: there may be 2 named groups called "path_type1" and
        "path_type2". Each named group represents one type of "path" and only
        1 of each type can occur in any given string. However, they are
        intrisically still considered the same thing: a path. This parameter
        allows you to specify that "path_type1" and "path_type2" should
        be converted to "path" by passing
        `extract_group(..., combine={"path": ["path_type1", "path_type2"], ...})`
        in the final dictionary. However, if more than 1 of any of the 2
        possible groups appear, a ValueError will be raised.
        
    Raises
    ------
    ValueError
        If more than 1 named group in a group of related named groups are found,
        this error is raised.
        OR If an invalid value is given to none_str.
    
    Returns
    -------
    Dict[str, Optional[str]]
        A dictionary of named groups and their associated values.
    """
    if none_str not in {"ignore", "to_none", "to_str"}:
        raise ValueError(f"Invalid none_str: {none_str}")
    
    def null_convert(input: t.Optional[str]) -> t.Optional[str]:
        if (input is not None and input != "") or none_str == "ignore":
            return input
        elif none_str == "to_none":
            return None
        else:
            return ""
    
    combine: t.Mapping[str, t.List[str]] = {} if combine is None else combine
    result = {}
    for name in names:
        related_name_groups = combine.get(name)
        related_name_groups = (
            [name]
            if related_name_groups is None
            else related_name_groups
        )
        for related in related_name_groups:
            try:
                group_value = matches.group(related)
            except IndexError:
                group_value = None
            group_value = null_convert(group_value)
            current_value = result.get(name)
            if current_value is not None and group_value is not None:
                raise ValueError(
                    f"'{name}' has already been filled by '{result.get(name)}' "
                    f"but a new value ('{group_value}') was found."
                )
            elif current_value is None:
                result[name] = group_value
        if name not in result:
            result[name] = None if none_str == "to_str" else ""
    return result


def make_parser(
    regex: re.Pattern[str],
    names: t.Iterable[str]
) -> t.Callable[[str, t.Mapping[t.str, t.Any]], t.Dict[str, t.Optional[str]]]:
    """
    Create a function that can parse locators.
    
    Parameters
    ----------
    regex: re.Pattern[str]
        A compiled regex pattern. You can create such an object by calling
        `re.compile` on a regex string.
    
    names: Iterable[str]
        A list of named groups that the parser should encounter in the regex
        matches.
    
    Returns
    -------
    Callable[[str, Mapping[t.Str, Any]], Dict[str, Optional[str]]]
        The parser function.
    """
    
    def parser(input: str, **kwargs: t.Any) -> t.Dict[str, t.Optional[str]]:
        matches = regex.search(input)
        if matches is None:
            return {}
        else:
            return extract_groups(matches, names, **kwargs)
    
    return parser


UNRESERVED: str = r"[0-9a-zA-Z._~\-]"
PERCENT_ENCODED: str = r"%[0-9a-fA-F]{2}"
SUB_DELIMS: str = r"[!\$&'\(\)\*\+\,\;\=]"
CHAR: str = f"{UNRESERVED}|{PERCENT_ENCODED}|{SUB_DELIMS}"


# Taken from validators.domain.pattern
DOMAIN_REGEX_RAW: str = (
    r"(?P<domain>"
    r"(?:[a-zA-Z0-9]"  # First character of the domain
    r"(?:[a-zA-Z0-9-_]{0,61}[A-Za-z0-9])?\.)+"  # Sub domain + hostname
    r"[A-Za-z0-9][A-Za-z0-9-_]{0,61}"  # First 61 characters of the gTLD
    r"[A-Za-z]"  # Last character of the gTLD
    r")"
)

# Taken from https://ihateregex.io/expr/ip/
IPV4_REGEX_RAW: str = (
    r"(?P<ipv4>"
    r"(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}"
    r")"
)

# Taken from https://stackoverflow.com/a/17871737
IPV6_REGEX_RAW: str = (
    r"(?P<ipv6>"
    r"([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|"          # 1:2:3:4:5:6:7:8
    r"([0-9a-fA-F]{1,4}:){1,7}:|"                         # 1::                              1:2:3:4:5:6:7::
    r"([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|"         # 1::8             1:2:3:4:5:6::8  1:2:3:4:5:6::8
    r"([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|"  # 1::7:8           1:2:3:4:5::7:8  1:2:3:4:5::8
    r"([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|"  # 1::6:7:8         1:2:3:4::6:7:8  1:2:3:4::8
    r"([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|"  # 1::5:6:7:8       1:2:3::5:6:7:8  1:2:3::8
    r"([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|"  # 1::4:5:6:7:8     1:2::4:5:6:7:8  1:2::8
    r"[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|"       # 1::3:4:5:6:7:8   1::3:4:5:6:7:8  1::8  
    r":((:[0-9a-fA-F]{1,4}){1,7}|:)|"                     # ::2:3:4:5:6:7:8  ::2:3:4:5:6:7:8 ::8       ::     
    r"fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|"     # fe80::7:8%eth0   fe80::7:8%1     (link-local IPv6 addresses with zone index)
    r"::(ffff(:0{1,4}){0,1}:){0,1}"
    r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
    r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|"          # ::255.255.255.255   ::ffff:255.255.255.255  ::ffff:0:255.255.255.255  (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
    r"([0-9a-fA-F]{1,4}:){1,4}:"
    r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
    r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])"           # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33 (IPv4-Embedded IPv6 Address)
    r")"
)


HOST_REGEX_RAW: str = f"(?P<host>{DOMAIN_REGEX_RAW}|{IPV4_REGEX_RAW}|{IPV6_REGEX_RAW})"


USERNAME_REGEX_RAW: str = f"(?P<username>({CHAR})+)"
PASSWORD_REGEX_RAW: str = f"(?P<password>({CHAR}|[:])+)"


USERINFO_REGEX_RAW: str = f"{USERNAME_REGEX_RAW}(:{PASSWORD_REGEX_RAW})?"


# 0 to 65535
PORT_REGEX_RAW: str = (
    r"(?P<port>"
    r"(6553[0-5]|655[0-2]\d|65[0-4]\d\d|6[0-4]\d\d\d|[1-5]?\d\d\d\d|[1-9]\d{0,3}|[0-9])"
    r")"
)


AUTHORITY_REGEX_RAW: str = (
    r"(?P<authority>"
    f"({USERINFO_REGEX_RAW}@)?({HOST_REGEX_RAW})(:{PORT_REGEX_RAW})?"
    r")"
)
AUTHORITY_REGEX: re.Pattern[str] = re.compile(f"^{AUTHORITY_REGEX_RAW}$")


SCHEME_REGEX_RAW: str = r"(?P<scheme>[a-zA-Z][a-zA-Z0-9+.-]*)"
PATH_INTERNAL_REGEX_RAW: str = f"(({CHAR}|[@])+(/({CHAR}|[@])+)*)/?|/?"
PATH_REGEX_RAW: str = r"(?P<path>" + PATH_INTERNAL_REGEX_RAW + r")"
SCP_PATH_REGEX_RAW: str = r"(?P<path>/?" + PATH_INTERNAL_REGEX_RAW + r")"
QUERY_REGEX_RAW: str = f"(?P<query>({CHAR}|[/\?])*)"
FRAGMENT_REGEX_RAW: str = f"(?P<fragment>({CHAR}|[/\?])*)"


FULL_URL_REGEX_RAW: str = (
    r"(?P<full_url>"
    f"{SCHEME_REGEX_RAW}://{AUTHORITY_REGEX_RAW}(/{PATH_REGEX_RAW})?(\?{QUERY_REGEX_RAW})?(#{FRAGMENT_REGEX_RAW})?"
    r")"
)
FULL_URL_REGEX: re.Pattern[str] = re.compile(f"^{FULL_URL_REGEX_RAW}$")


DIRTY_URL_REGEX_RAW: str = (
    r"(?P<dirty_url>"
    f"({SCHEME_REGEX_RAW}://)?"
    f"({AUTHORITY_REGEX_RAW})?"
    f"(/?{PATH_REGEX_RAW}?)?"
    f"(\?{QUERY_REGEX_RAW})?"
    f"(#{FRAGMENT_REGEX_RAW})?"
    r")"
)
DIRTY_URL_REGEX: re.Pattern[str] = re.compile(f"^{DIRTY_URL_REGEX_RAW}$")


SCP_FULL_LOC_REGEX_RAW: str = (
    r"(?P<scp_full>"
    f"{AUTHORITY_REGEX_RAW}(:{SCP_PATH_REGEX_RAW})?"
    r")"
)
SCP_FULL_LOC_REGEX: re.Pattern[str] = re.compile(f"^{SCP_FULL_LOC_REGEX_RAW}$")

SCP_DIRTY_LOC_REGEX_RAW: str = (
    r"(?P<scp_dirty>"
    f"({AUTHORITY_REGEX_RAW})?(:{SCP_PATH_REGEX_RAW})?"
    r")"
)
SCP_DIRTY_LOC_REGEX: re.Pattern[str] = re.compile(f"^{SCP_DIRTY_LOC_REGEX_RAW}$")


parse_authority = make_parser(
    AUTHORITY_REGEX,
    ["authority", "host", "domain", "ipv4", "ipv6", "username", "password", "port"]
)
"""
Parse the authority segment in a URL and split it into its consituent parts.

Parameters
----------
authority: str
    The authority segment in a URL.

**kwargs: Any
    Keyword arguments to be passed to `extract_groups`.

Returns
-------
Dict[str, Optional[str]]
    The parts of the authority segment.
"""


parse_full_url = make_parser(
    FULL_URL_REGEX, [
        "full_url",
        "scheme",
        "authority",
        "host",
        "domain",
        "ipv4",
        "ipv6",
        "username",
        "password",
        "port",
        "path",
        "query",
        "fragment"
    ]
)
"""
Parse a full URL and split it into its constituent parts.

Parameters
----------
url: str
    The full URL.

**kwargs: Any
    Keyword arguments to be passed to `extract_groups`.

Returns
-------
Dict[str, Optional[str]]
    The parts of the URL.
"""


parse_dirty_url = make_parser(
    DIRTY_URL_REGEX,
    [
        "dirty_url",
        "scheme",
        "authority",
        "host",
        "domain",
        "ipv4",
        "ipv6",
        "username",
        "password",
        "port",
        "path",
        "query",
        "fragment"
    ]
)
"""
Parse a URL (assuming that it was inputted by a lazy human) and split it
into its constituent parts.

Parameters
----------
url: str
    The user-inputted URL.

**kwargs: Any
    Keyword arguments to be passed to `extract_groups`.

Returns
-------
Dict[str, Optional[str]]
    The parts of the URL.
"""


parse_scp_full_loc = make_parser(
    SCP_FULL_LOC_REGEX,
    [
        "scp_full",
        "authority",
        "host",
        "domain",
        "ipv4",
        "ipv6",
        "username",
        "password",
        "path"
    ]
)
"""
Parse a full SCP (secure copy) locator and split it into its constituent parts.

Parameters
----------
url: str
    The full URL.

**kwargs: Any
    Keyword arguments to be passed to `extract_groups`.

Returns
-------
Dict[str, Optional[str]]
    The parts of the locator.
"""


parse_scp_dirty_loc = make_parser(
    SCP_DIRTY_LOC_REGEX,
    [
        "scp_dirty",
        "authority",
        "host",
        "domain",
        "ipv4",
        "ipv6",
        "username",
        "password",
        "path"
    ]
)
"""
Parse a SCP locator (assuming that it was inputted by a lazy human) and split it
into its constituent parts.

Parameters
----------
url: str
    The user-inputted URL.

**kwargs: Any
    Keyword arguments to be passed to `extract_groups`.

Returns
-------
Dict[str, Optional[str]]
    The parts of the locator.
"""



from __future__ import annotations
import typing as t

from .parser import (
    parse_authority,
    parse_dirty_url,
    parse_full_url,
    parse_scp_dirty_loc
)

__all__ = [
    "BaseLocator",
    "UniformResourceLocator",
    "DirtyLocator",
    "LocatorBuilder",
    "UrlAuthority"
]


LOCATOR_FIELDS: str = """
Fields
------
scheme: str
    The scheme used to access the remote repository.
    Comes before everything else in a URL and is suffixed by '://'.
    Examples: https, ssh

host: str
    The host where the remote repository is located.
    Examples: github.com, 1.1.1.1

username: str
    The username used to access the remote repository.
    Comes before the host and is separated from the latter by '@'.
    Examples: git (in git@github.com)

password: str
    Password of the user used to authenticate themselves.
    If present, must come after a username (separated by ':') and before
    a host name (separated by '@').
    RFC 3986 recommends that you don't ever use this because it's a huge
    security vulnerability.
    However, this is still used by github when you use HTTPS.

port: str
    The port used to connect to the server.
    If present, must come after the host name and must be an integer
    between 0 and 65535 (inclusive).

path: str
    The path to the remote repository
    (not referring to where the local clone is located).
    Comes after the host and separated from it by '/'.
    Examples: RenoirTan/QuickClone
    (in https://github.com/RenoirTan/QuickClone)

query: str
    The parameters in the URL.
    Comes after the path or host and the start of the query section is
    denoted by '?'.
    You can chain multiple key-value pairs using '&' as the separator.
    Examples: key=value (in https://example.com?key=value)

fragment: str
    The fragment identifying a secondary resource.
    The last part of the URL. The fragment section is denoted by '#'.
    Examples: History
    (in https://en.wikipedia.org/wiki/Python_(programming_language)#History)
"""


FIELDS = ["scheme", "host", "username", "password", "port", "path", "query", "fragment"]


class BaseLocator(object):
    """
    Base class for classes representing Uniform Resource Locators (URLs).
    This stores the parts of the URL like the domain name or path in the URL.
    
    Fields
    ------
    scheme: str
        The scheme used to access the remote repository.
        Comes before everything else in a URL and is suffixed by '://'.
        Examples: https, ssh

    host: str
        The host where the remote repository is located.
        Examples: github.com, 1.1.1.1

    username: str
        The username used to access the remote repository.
        Comes before the host and is separated from the latter by '@'.
        Examples: git (in git@github.com)

    password: str
        Password of the user used to authenticate themselves.
        If present, must come after a username (separated by ':') and before
        a host name (separated by '@').
        RFC 3986 recommends that you don't ever use this because it's a huge
        security vulnerability.
        However, this is still used by github when you use HTTPS.
    
    port: str
        The port used to connect to the server.
        If present, must come after the host name and must be an integer
        between 0 and 65535 (inclusive).

    path: str
        The path to the remote repository
        (not referring to where the local clone is located).
        Comes after the host and separated from it by '/'.
        Examples: RenoirTan/QuickClone
        (in https://github.com/RenoirTan/QuickClone)

    query: str
        The parameters in the URL.
        Comes after the path or host and the start of the query section is
        denoted by '?'.
        You can chain multiple key-value pairs using '&' as the separator.
        Examples: key=value (in https://example.com?key=value)

    fragment: str
        The fragment identifying a secondary resource.
        The last part of the URL. The fragment section is denoted by '#'.
        Examples: History
        (in https://en.wikipedia.org/wiki/Python_(programming_language)#History)
    
    kwargs: Dict[str, Any]
        Miscellaneous data.
    """
    
    def __init__(
        self,
        scheme: str = "",
        host: str = "",
        username: str = "",
        password: str = "",
        path: str = "",
        port: str = "",
        query: str = "",
        fragment: str = "",
        **kwargs
    ) -> None:
        """
        Initiate an object representing a URL.
        
        Parameters
        ----------
        self: BaseLocator
            The object to be initialised.
        
        scheme: str = ""
            The scheme used to access the remote repository.
            Comes before everything else in a URL and is suffixed by '://'.
            Examples: https, ssh

        host: str = ""
            The host where the remote repository is located.
            Examples: github.com, 1.1.1.1

        username: str = ""
            The username used to access the remote repository.
            Comes before the host and is separated from the latter by '@'.
            Examples: git (in git@github.com)

        password: str = ""
            Password of the user used to authenticate themselves.
            If present, must come after a username (separated by ':') and before
            a host name (separated by '@').
            RFC 3986 recommends that you don't ever use this because it's a huge
            security vulnerability.
            However, this is still used by github when you use HTTPS.
        
        port: str = ""
            The port used to connect to the server.
            If present, must come after the host name and must be an integer
            between 0 and 65535 (inclusive).

        path: str = ""
            The path to the remote repository
            (not referring to where the local clone is located).
            Comes after the host and separated from it by '/'.
            Examples: RenoirTan/QuickClone
            (in https://github.com/RenoirTan/QuickClone)

        query: str = ""
            The parameters in the URL.
            Comes after the path or host and the start of the query section is
            denoted by '?'.
            You can chain multiple key-value pairs using '&' as the separator.
            Examples: key=value (in https://example.com?key=value)

        fragment: str = ""
            The fragment identifying a secondary resource.
            The last part of the URL. The fragment section is denoted by '#'.
            Examples: History
            (in https://en.wikipedia.org/wiki/Python_(programming_language)#History)
        
        **kwargs: Any
            Miscellaneous data.
        """
        self.scheme = scheme
        self.host = host
        self.username = username
        self.password = password
        self.port = port
        self.path = path
        self.query = query
        self.fragment = fragment
        self.kwargs = kwargs
    
    def __str__(self) -> str:
        scheme_part = f"{self.get_scheme()}://" if self.get_scheme() != "" else ""
        authority_part = str(UrlAuthority(
            self.get_host(),
            self.get_username(),
            self.get_password(),
            self.get_port()
        ))
        path_part = f"/{self.get_path()}" if self.get_path() != "" else ""
        query_part = f"?{self.get_query()}" if self.get_query() != "" else ""
        fragment_part = f"#{self.get_fragment()}" if self.get_fragment() != "" else ""
        return scheme_part + authority_part + path_part + query_part + fragment_part
        
    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"scheme={repr(self.scheme)}, "
            f"host={repr(self.host)}, "
            f"username={repr(self.username)}, "
            f"password={repr(self.password)}, "
            f"port={repr(self.port)}, "
            f"path={repr(self.path)}, "
            f"query={repr(self.query)}, "
            f"fragment={repr(self.fragment)})"
        )
    
    def __iter__(self) -> t.Generator[t.Tuple[str, str], None, None]:
        yield "scheme", self.get_scheme()
        yield "host", self.get_host()
        yield "username", self.get_username()
        yield "password", self.get_password()
        yield "port", self.get_port()
        yield "path", self.get_path()
        yield "query", self.get_query()
        yield "fragment", self.get_fragment()
        yield from self.kwargs.items()
    
    def get_scheme(self) -> str:
        """Get the scheme used to access the remote repository."""
        return self.scheme

    def get_host(self) -> str:
        """Get the host name of the website hosting the remote repository."""
        return self.host

    def get_username(self) -> str:
        """Get the username used to access the remote repository."""
        return self.username

    def get_password(self) -> str:
        """Get the password used to access the remote repository."""
        return self.password
    
    def get_port(self) -> str:
        """
        Get the port used to access the server hosting the remote repository.
        """
        return self.port

    def get_path(self) -> str:
        """Get the path of the remote repository (in the website)."""
        return self.path

    def get_query(self) -> str:
        """Get the parameters (as a single string)."""
        return self.query

    def get_fragment(self) -> str:
        """Get the fragment."""
        return self.fragment


class UniformResourceLocator(BaseLocator):
    """
    A full uniform resource locator (hopefully RFC 3986 compliant).
    
    Fields
    ------
    scheme: str
        The scheme used to access the remote repository.
        Comes before everything else in a URL and is suffixed by '://'.
        Examples: https, ssh

    host: str
        The host where the remote repository is located.
        Examples: github.com, 1.1.1.1

    username: str
        The username used to access the remote repository.
        Comes before the host and is separated from the latter by '@'.
        Examples: git (in git@github.com)

    password: str
        Password of the user used to authenticate themselves.
        If present, must come after a username (separated by ':') and before
        a host name (separated by '@').
        RFC 3986 recommends that you don't ever use this because it's a huge
        security vulnerability.
        However, this is still used by github when you use HTTPS.
    
    port: str
        The port used to connect to the server.
        If present, must come after the host name and must be an integer
        between 0 and 65535 (inclusive).

    path: str
        The path to the remote repository
        (not referring to where the local clone is located).
        Comes after the host and separated from it by '/'.
        Examples: RenoirTan/QuickClone
        (in https://github.com/RenoirTan/QuickClone)

    query: str
        The parameters in the URL.
        Comes after the path or host and the start of the query section is
        denoted by '?'.
        You can chain multiple key-value pairs using '&' as the separator.
        Examples: key=value (in https://example.com?key=value)

    fragment: str
        The fragment identifying a secondary resource.
        The last part of the URL. The fragment section is denoted by '#'.
        Examples: History
        (in https://en.wikipedia.org/wiki/Python_(programming_language)#History)
    """
    
    def __str__(self) -> str:
        for fault in self.find_faults():
            raise fault
        return super().__str__()
    
    def find_faults(self) -> t.Generator[BaseException, None, None]:
        """
        Find errors in this URL. This method returns a generator object, which
        allows you to iterate over this method.
        
        Yields
        ------
        BaseException
            All the errors that are encountered when reviewing the URL. If no
            errors are found, no exceptions will be yielded.
        """
        if self.get_username() == "" and self.get_password() != "":
            yield ValueError("Username not given but password given.")
        if self.get_scheme() == "":
            yield ValueError("No scheme specified.")
        if self.get_host() == "":
            yield ValueError("Empty host.")

    def validate(self) -> bool:
        """
        Validate this URL by making sure no errors are found in this URL.
        This uses the `find_faults` method to try and find errors.
        
        Returns
        -------
        bool
            `True` if this URL has no errors in it and `False` otherwise.
        """
        return sum(map(lambda _: 1, self.find_faults())) == 0

    def detect_explicitness(self, force_scp: bool, ignore_force_scp: bool) -> None:
        """
        Detect whether this locator was explicitly written as an SCP locator
        and set the `self.kwargs["explicit_scp"]` value to `True`.
        
        Parameters
        ----------
        configs: SmartConfigurator
            The user's configuration.
        
        ignore_force_scp:
            Whether to ignore the `options.remote.force_scp` configuration.
        """
        if force_scp and not ignore_force_scp:
            self.kwargs["explicit_scp"] = True
    
    @classmethod
    def process_url(cls, url: str) -> UniformResourceLocator:
        """
        Process a URL and turn it into `UniformResourceLocator` object. The
        URL must have at least a scheme and host.
        
        Parameters
        ----------
        cls: Type[UniformResourceLocator]
            The UniformResourceLocator class.
        
        url: str
            The URL to be parsed and decomposed into a UniformResourceLocator.
        
        Raises
        ------
        ValueError
            If an error was found in the URL. Only the first error encountered
            will be raised.
        
        Returns
        -------
        UniformResourceLocator
        """
        parsed = parse_full_url(url, none_str="to_str")
        if parsed == {}:
            raise ValueError(f"Invalid URL: {url}")
        url: UniformResourceLocator = UniformResourceLocator(**parsed)
        for fault in url.find_faults():
            raise fault
        return url
    
    @classmethod
    def from_user_and_defaults(
        cls,
        user_input: DirtyLocator,
        defaults: LocatorBuilder
    ) -> UniformResourceLocator:
        """
        Create a final URL from the user input and defaults drawn from user
        configurations or hard coded into QuickClone.
        
        Parameters
        ----------
        cls: Type[UniformResourceLocator]
            The UniformResourceLocator class.
        
        user_input: DirtyLocator
            The user-inputted URL.
        
        defaults: LocatorBuilder
            The defaults from configs.
        
        Raises
        ------
        ValueError
            If vital bits are missing.
        
        Returns
        -------
        UniformResourceLocator
        """
        parts = dict(defaults)
        for key, value in user_input:
            if value != "":
                parts[key] = value
        url = cls(**parts)
        for fault in url.find_faults():
            raise fault
        return url


class DirtyLocator(BaseLocator):
    """
    A locator object representing a user-inputted URL (with parts possibly
    missing).
    
    Fields
    ------
    scheme: str
        The scheme used to access the remote repository.
        Comes before everything else in a URL and is suffixed by '://'.
        Examples: https, ssh

    host: str
        The host where the remote repository is located.
        Examples: github.com, 1.1.1.1

    username: str
        The username used to access the remote repository.
        Comes before the host and is separated from the latter by '@'.
        Examples: git (in git@github.com)

    password: str
        Password of the user used to authenticate themselves.
        If present, must come after a username (separated by ':') and before
        a host name (separated by '@').
        RFC 3986 recommends that you don't ever use this because it's a huge
        security vulnerability.
        However, this is still used by github when you use HTTPS.
    
    port: str
        The port used to connect to the server.
        If present, must come after the host name and must be an integer
        between 0 and 65535 (inclusive).

    path: str
        The path to the remote repository
        (not referring to where the local clone is located).
        Comes after the host and separated from it by '/'.
        Examples: RenoirTan/QuickClone
        (in https://github.com/RenoirTan/QuickClone)

    query: str
        The parameters in the URL.
        Comes after the path or host and the start of the query section is
        denoted by '?'.
        You can chain multiple key-value pairs using '&' as the separator.
        Examples: key=value (in https://example.com?key=value)

    fragment: str
        The fragment identifying a secondary resource.
        The last part of the URL. The fragment section is denoted by '#'.
        Examples: History
        (in https://en.wikipedia.org/wiki/Python_(programming_language)#History)
    """
    
    @classmethod
    def process_dirty_url(cls, dirty_url: str) -> DirtyLocator:
        """
        Process a user-inputted URL (`dirty_url`).

        Parameters
        ----------
        cls: typing.Type[UrlAuthority]
            The UrlAuthority class.

        dirty_url: str
            The user-inputted URL.
        
        Raises
        ------
        ValueError
            If regex could not find matches using the `dirty_url`.

        Returns
        -------
        UrlAuthority
        """
        result: t.Dict[str, str] = parse_dirty_url(dirty_url, none_str="to_str")
        if result == {}:
            result = parse_scp_dirty_loc(dirty_url, none_str="to_str")
            result["scheme"] = "ssh"
            result["explicit_scp"] = True
        if result == {}:
            raise ValueError(f"Could not match {dirty_url}")
        return cls(**result)


class LocatorBuilder(BaseLocator):
    """
    A builder which stores user/program specified defaults to generate valid
    URLs from user inputs.
    
    Fields
    ------
    scheme: str
        The scheme used to access the remote repository.
        Comes before everything else in a URL and is suffixed by '://'.
        Examples: https, ssh

    host: str
        The host where the remote repository is located.
        Examples: github.com, 1.1.1.1

    username: str
        The username used to access the remote repository.
        Comes before the host and is separated from the latter by '@'.
        Examples: git (in git@github.com)

    password: str
        Password of the user used to authenticate themselves.
        If present, must come after a username (separated by ':') and before
        a host name (separated by '@').
        RFC 3986 recommends that you don't ever use this because it's a huge
        security vulnerability.
        However, this is still used by github when you use HTTPS.
    
    port: str
        The port used to connect to the server.
        If present, must come after the host name and must be an integer
        between 0 and 65535 (inclusive).

    path: str
        The path to the remote repository
        (not referring to where the local clone is located).
        Comes after the host and separated from it by '/'.
        Examples: RenoirTan/QuickClone
        (in https://github.com/RenoirTan/QuickClone)

    query: str
        The parameters in the URL.
        Comes after the path or host and the start of the query section is
        denoted by '?'.
        You can chain multiple key-value pairs using '&' as the separator.
        Examples: key=value (in https://example.com?key=value)

    fragment: str
        The fragment identifying a secondary resource.
        The last part of the URL. The fragment section is denoted by '#'.
        Examples: History
        (in https://en.wikipedia.org/wiki/Python_(programming_language)#History)
    """

    DEFAULT_SCHEME: str = "https"
    """
    The default scheme used to access the remote repository. By default this is "https".
    """

    DEFAULT_HOST: str = "github.com"
    """
    The default host where the remote repository is located. By default this is "github.com".
    """

    def get_scheme(self) -> str:
        """Get the scheme used to access the remote repository."""
        return self.DEFAULT_SCHEME if self.scheme == "" else self.scheme

    def get_host(self) -> str:
        """Get the host name of the website hosting the remote repository."""
        return self.DEFAULT_HOST if self.host == "" else self.host


class UrlAuthority(object):
    """
    The authority component in a URL. This includes the host name, username
    password and port in the URL. Example: username:password@example.com:80

    Parameters
    ----------
    self: UrlAuthority
        The UrlAuthority object to be initialised.

    host: str
        The host name of the URL.

    username: str = ""
        The username in the URL.

    password: str = ""
        The password in the URL.
    
    port: str = ""
        The port in the URL. Must be a value between 0 and 65535 (inclusive).
    """

    def __init__(
        self,
        host: str,
        username: str = "",
        password: str = "",
        port: str = ""
    ) -> None:
        self.host = host
        self.username = username
        self.password = password
        self.port = port
    
    def __str__(self) -> str:
        userpass_part = self.username
        if self.password != "":
            userpass_part += f":{self.password}"
        if len(userpass_part) > 0:
            userpass_part += "@"
        port_part = f":{self.port}" if self.port != "" else ""
        return userpass_part + self.host + port_part

    @classmethod
    def process_authority(cls, authority: str) -> UrlAuthority:
        """
        Convert an authority (as a string) into a `UrlAuthority` object.

        Parameters
        ----------
        cls: typing.Type[UrlAuthority]
            The UrlAuthority class.

        authority: str
            The authority section in the URL.
        
        Raises
        ------
        ValueError
            If password provided without providing username.

        Returns
        -------
        UrlAuthority
        """
        result = parse_authority(authority, none_str="to_str")
        host = result["host"]
        username = result["username"]
        password = result["password"]
        port = result["port"]
        if username == "" and password != "":
            raise ValueError("Password provided without username.")
        return cls(host, username, password, port)



from pathlib import Path
import shutil
import subprocess

from quickclone.config.common import USER_CACHE_FOLDER, CACHE_ITEMS

USER_CACHE_FOLDER_OLD = Path.home() / ".cache" / "quicktoml"
"""
Path to the incorrectly named cache folder: `~/.cache/quicktoml`
"""

def quickclone_cache_move() -> int:
    """
    Fix for v4.0 or below.
    
    Prompt user to move or copy cache from `~/.cache/quicktoml` to
    `~/.cache/quickclone`, then perform an operation based on the user's choice.
    See #Returns for what operations are available.
    The main logic of this function should only be activated if
    `~/.cache/quickclone` is not detected and `~/.cache/quicktoml` is non-empty.
    
    Returns
    -------
    status: int
        The status code indicates what operation was performed.
        0: Nothing had to be done, the cache folder is already at
            `~/.cache/quickclone` or if `~/.cache/quicktoml` was not detected.
        1: User didn't want to move `~/.cache/quicktoml` to
            `~/.cache/quickclone`.
        2: User moved `~/.cache/quicktoml` to `~/.cache/quickclone`.
        3: User copied `~/.cache/quicktoml` to `~/.cache/quickclone`.
    """
    # if ~/.cache/quickclone already exists
    if USER_CACHE_FOLDER.exists():
        return 0
    # detect useful ~/.cache/quicktoml
    if USER_CACHE_FOLDER_OLD.exists():
        if not USER_CACHE_FOLDER_OLD.is_dir():
            return 0
        # if ~/.cache/quicktoml empty
        try:
            next(USER_CACHE_FOLDER_OLD.iterdir())
        except StopIteration:
            return 0
    else:
        return
    print("⚠⚠⚠ WARNING ⚠⚠⚠")
    print(
        f"I made a mistake when writing the path of the cache folder, it was supposed to be at "
        f"'{USER_CACHE_FOLDER}' but instead I wrote '{USER_CACHE_FOLDER_OLD}'.",
        end="\n\n"
    )
    print(
        f"You can choose to either MOVE or COPY the contents of '{USER_CACHE_FOLDER_OLD}' "
        f"to '{USER_CACHE_FOLDER}'.",
        end="\n\n"
    )
    print(
        f"If you decide to move ALL of the contents of the cache "
        f"folder, any program that was using '{USER_CACHE_FOLDER_OLD}' cannot access their data at "
        f"that location. However, from what I know, no program other than QuickClone version 0.4.0 "
        f"and below uses this folder, so moving the cache folder should be safe.",
        end="\n\n"
    )
    print(
        f"If you choose to copy the cache from '{USER_CACHE_FOLDER_OLD}' to '{USER_CACHE_FOLDER}' "
        f"any program using '{USER_CACHE_FOLDER_OLD}' would have their data preserved, but this "
        f"would leave 2 copies of the same files inside each folder.",
        end="\n\n"
    )
    print("SELECT AN OPTION")
    print(f"1. MOVE contents of '{USER_CACHE_FOLDER_OLD}' to '{USER_CACHE_FOLDER}'.")
    print(f"2. COPY contents of '{USER_CACHE_FOLDER_OLD}' to '{USER_CACHE_FOLDER}'.")
    print(f"3. CANCEL")
    # prompt user whether to move directories
    answer = 0
    while True:
        answer = input("[1/2/3] ==> ")
        if answer in {"1", "2", "3"}:
            answer = int(answer)
            break
        else:
            print(f"Invalid answer: {answer}")
    ACTION = [_move_cache_folder, _copy_cache_folder, _exit_cache_move]
    return ACTION[answer-1]()


def _move_cache_folder() -> int:
    shutil.move(USER_CACHE_FOLDER_OLD, USER_CACHE_FOLDER)
    return 2

def _copy_cache_folder() -> int:
    subprocess.run(["mkdir", "-p", str(USER_CACHE_FOLDER)])
    for filename in CACHE_ITEMS:
        old = USER_CACHE_FOLDER_OLD / filename
        new = USER_CACHE_FOLDER / filename
        if old.exists():
            shutil.copy2(old, new)
    return 3
    
def _exit_cache_move() -> int:
    return 1



import argparse
from pathlib import Path
import typing as t
import subprocess
import sys

from quickclone import DESCRIPTION, NAME, VERSION
from quickclone.compatibility import v0_4_0
from quickclone.config.cache import (
    load_caches,
    dump_caches,
    get_cache_value,
    set_cache_value,
    AVAILABLE_CACHES
)
from quickclone.config.common import DEFAULTS_FOLDER
from quickclone.config.configurator import load_user_config
from quickclone.delegation.tasks import create_clone_command
from quickclone.delegation.vcs.common import Command
from quickclone.remote import DirtyLocator, UniformResourceLocator, UrlAuthority


def program():
    sys.exit(main(sys.argv))


def create_argument_parser() -> argparse.ArgumentParser:
    app = argparse.ArgumentParser(description=f"{NAME} v{VERSION}: {DESCRIPTION}")
    app.add_argument(
        "--version",
        "-v",
        dest="show_version",
        help="display the version number and exit",
        action="store_const",
        const=True,
        default=False
    )
    app.add_argument(
        "remote_url",
        metavar="REMOTE_URL",
        nargs="?",
        type=str,
        default="",
        help="the url of the remote to be cloned"
    )
    app.add_argument(
        "dest_path",
        metavar="DEST_PATH",
        nargs="?",
        type=str,
        default="",
        help="the directory where the remote repository should be cloned to"
    )
    app.add_argument(
        "--last-clone",
        "-L",
        dest="get_last_clone",
        action="store_const",
        const=True,
        default=False,
        help="get the directory of where the last repo was cloned to"
    )
    app.add_argument(
        "--pretend",
        "-P",
        dest="pretend",
        action="store_const",
        const=True,
        default=False,
        help=(
            "if this flag is found, QuickClone will not perform any important "
            "actions that it would have if this flag is not found"
        )
    )
    app.add_argument(
        "--system",
        "-S",
        dest="vcs",
        help="which version control system to use: git, mercurial/hg"
    )
    app.add_argument(
        "--config-file",
        "-C",
        dest="config_file",
        metavar="CONFIG_FILE_PATH",
        help="override the default config file location"
    )
    app.add_argument(
        "--ignore",
        "-I",
        dest="ignore",
        metavar="CONFIG_KEY",
        action="append",
        required=False,
        default=[],
        help=(
            "what part of the config file to ignore. "
            "you can use dot-separated keys (like 'options.local.remotes_dir') or "
            "their short forms to specify which config values to ignore"
        )
    )
    app.add_argument(
        "--test",
        "-T",
        dest="tests",
        metavar="TEST",
        action="append",
        required=False,
        default=[],
        help=(
            "which tests to conduct: "
            "parse_authority, parse_full_url, parse_dirty_url, print_defaults_folder, config_file"
        )
    )
    return app


def ignore_config(keys: t.List[str]) -> t.Set[str]:
    SHORT_FORMS = {
        "d": "options.local.remotes_dir",
        "s": "options.remote.force_scp"
    }
    ignored = set(keys)
    for short, long in SHORT_FORMS.items():
        if short in keys:
            ignored.add(long)
    return ignored


def do_compatibility():
    compat_status = v0_4_0.quickclone_cache_move()
    if compat_status == 1:
        print("Cancelled move/copy!")


def main(argv: t.List[str]) -> int:
    do_compatibility()
    load_caches(AVAILABLE_CACHES)
    app = create_argument_parser()
    args = app.parse_args(argv[1:])
    if args.show_version:
        print(f"{NAME} v{VERSION}")
        return 0
    elif args.get_last_clone:
        last_clone: t.Optional[str] = get_cache_value("last_clone")
        if last_clone is not None and last_clone != "":
            print(last_clone)
        return 0
    if len(args.tests) > 0:
        successes, test_count = conduct_tests(args.tests, args.remote_url)
        if successes < test_count:
            print("Not all tests succeeded")
            dump_caches(AVAILABLE_CACHES)
            return 1
        else:
            dump_caches(AVAILABLE_CACHES)
            return 0
    else:
        return normal(args)


# Call this function if quickclone is run with the normal set of clargs.
def normal(args: argparse.Namespace) -> int:
    dirty = DirtyLocator.process_dirty_url(args.remote_url)
    ignored = ignore_config(args.ignore)
    try:
        configs = load_user_config(None if args.config_file is None else Path(args.config_file))
        builder = configs.to_locator_builder()
        built_url = UniformResourceLocator.from_user_and_defaults(dirty, builder)
        vcs = configs.from_dotted_string("vcs.command")
        if args.vcs is not None:
            vcs = args.vcs
        clone_command = create_clone_command(
            vcs,
            configs,
            built_url,
            args.dest_path,
            [],
            {},
            ignored
        )
        print(f"Command> {clone_command.format_command_str()}")
        if args.pretend:
            print("pretend flag found! Not executing command.")
        else:
            run_command(clone_command)
    except Exception as e:
        dump_caches(AVAILABLE_CACHES)
        raise e
        print(f"Error occurred:\n{e}")
        return 1
    else:
        dump_caches(AVAILABLE_CACHES)
        return 0


def run_command(command: Command) -> subprocess.CompletedProcess:
    result = command.run()
    if isinstance(result, subprocess.CompletedProcess):
        set_cache_value("last_clone", command.dest_path)
        return result
    elif isinstance(result, subprocess.SubprocessError):
        raise result


def conduct_tests(tests: t.List[str], remote_url: str) -> t.Tuple[int, int]:
    success_counts: int = 0
    for test in tests:
        try:
            if test == "parse_authority":
                authority = UrlAuthority.process_authority(remote_url)
                print(authority)
                success_counts += 1
            elif test == "parse_full_url":
                url = UniformResourceLocator.process_url(remote_url)
                print(url)
                success_counts += 1
            elif test == "parse_dirty_url":
                dirty_url = DirtyLocator.process_dirty_url(remote_url)
                print(dirty_url)
                success_counts += 1
            elif test == "print_defaults_folder":
                print(DEFAULTS_FOLDER)
                success_counts += 1
            elif test == "config_file":
                success_counts += int(test_config_file())
            else:
                print(f"Unrecognised test: {test}")
        except:
            pass
    return success_counts, len(tests)


def test_config_file() -> bool:
    try:
        config = load_user_config()
        print(config)
        for key in ["options.remote.scheme"]:
            print(f"{key}: {config.from_dotted_string(key)}")
    except Exception as e:
        print(e)
        return False
    else:
        return True


if __name__ == "__main__":
    program()



import typing as t

from quickclone.config.configurator import SmartConfigurator
from quickclone.local import local_dest_path
from quickclone.remote import UniformResourceLocator, remote_to_string

from .errors import InvalidVcsError
from .vcs.common import Command
from .vcs.git import GitCloneCommand
from .vcs.mercurial import MercurialCloneCommand


def create_clone_command(
    vcs: str,
    configs: SmartConfigurator,
    built_url: t.Union[UniformResourceLocator, str],
    dest_path: str = "",
    cla_list: t.Optional[t.Iterable[str]] = None,
    cla_dict: t.Optional[t.Mapping[str, str]] = None,
    ignored: t.Optional[t.Set[str]] = None
) -> Command:
    """
    Create a clone command for a version control system.
    
    Parameters
    ----------
    vcs: str
        Which version control system to use.
    
    configs: SmartConfigurator
        Configuration object.
    
    built_url: UniformResourceLocator | str
        The locator of the remote repository.
    
    dest_path: str = ""
        The local destination path of the cloned repository. The value of this
        argument should come from the command line instead of being the
        post-processed result from other functions.
    
    cla_list: Iterable[str] | None = None
        Extra command line arguments stored as an iterable.
    
    cla_dict: Dict[str, str] | None = None
        Extra command line arguments sotred as a mapping object.
    
    ignored: Set[str] | None = None
        Set of config options to ignore.
    
    Raises
    ------
    InvalidVcsError
        If `vcs` is invalid or not supported.
    
    Returns
    -------
    Command
        The command used to clone the remote repository.
    """
    if cla_list is None:
        cla_list = list()
    if cla_dict is None:
        cla_dict = dict()
    if ignored is None:
        ignored = set()
    
    built_url.detect_explicitness(
        configs.from_dotted_string("options.remote.force_scp"),
        "options.remote.force_scp" in ignored
    )
    final_url = remote_to_string(built_url, vcs)
    dest_path = local_dest_path(
        dest_path,
        configs.from_dotted_string("options.local.remotes_dir"),
        built_url.get_host(),
        built_url.get_path(),
        "options.local.remotes_dir" in ignored
    )
    
    return create_clone_command_with_processed(
        vcs,
        configs,
        final_url,
        dest_path,
        cla_list,
        cla_dict,
        ignored
    )


def create_clone_command_with_processed(
    vcs: str,
    configs: SmartConfigurator,
    final_url: str,
    dest_path: str,
    cla_list: t.Iterable[str],
    cla_dict: t.Mapping[str, str],
    ignored: t.Set[str]
) -> Command:
    """
    Create a clone command with the valid processed values passed in.
    """
    if vcs in {"git"}:
        return GitCloneCommand(final_url, dest_path, *cla_list, **cla_dict)
    elif vcs in {"mercurial", "hg"}:
        return MercurialCloneCommand(final_url, dest_path, *cla_list, **cla_dict)
    else:
        raise InvalidVcsError(vcs)



from __future__ import annotations
from pathlib import Path
import typing as t
from urllib.parse import urlencode

import toml

from quickclone.remote.locators import LocatorBuilder
from .common import DEFAULTS_FOLDER, USER_CONFIG_FILE


class Configurator(object):
    """
    A wrapper around a configuration file meant for QuickClone. This class
    has special methods for retrieving nested items.
    """
    
    def __init__(self, configuration: t.Mapping[str, t.Any]) -> None:
        self.configuration = configuration
    
    def __getitem__(self, key: t.Union[str, t.Iterable[str]]) -> t.Any:
        if type(key) == str:
            return self.configuration.get(key)
        container = self.configuration
        for part in key:
            container = container.get(part)
            if container is None:
                return ""
        return container if container is not None else ""
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(configuration={self.configuration})"
    
    def from_dotted_string(self, key: str) -> t.Optional[t.Any]:
        """
        Retrieve an item using a dot-separated key.
        
        Parameters
        ----------
        key: str
            The dot-separated key to the item.
        
        Returns
        -------
        Any
            The item associated with the key.
        """
        return self[key.split(".")]
    
    def to_locator_builder(self) -> LocatorBuilder:
        """
        Create a `LocatorBuilder` from the configs stored in this object.
        
        Returns
        -------
        LocatorBuilder
            The final `LocatorBuilder` object.
        """
        scheme = self.from_dotted_string("options.remote.scheme")
        host = self.from_dotted_string("options.remote.host")
        username = self.from_dotted_string("options.remote.username")
        password = self.from_dotted_string("options.remote.password")
        if password != "":
            print("I might remove passwords because it's a safety hazard.")
        port = self.from_dotted_string("options.remote.port")
        base_path = self.from_dotted_string("options.remote.base_path")
        path = self.from_dotted_string("options.remote.path")
        query = self.from_dotted_string("options.remote.query")
        fragment = self.from_dotted_string("options.remote.fragment")
        if type(query) == dict:
            query = urlencode(query)
        return LocatorBuilder(
            scheme=scheme,
            host=host,
            username=username,
            password=password,
            port=port,
            base_path=base_path,
            path=path,
            query=query,
            fragment=fragment
        )
    
    @classmethod
    def from_file(cls, path: Path) -> Configurator:
        """
        Load the configuration from a file.
        
        Parameters
        ----------
        path: Path
            Path to the configuration file.
        
        Returns
        -------
        Configurator
            The configuration loaded from the file stored as a `Configurator`
            object.
        """
        configuration = toml.load(path)
        return cls(configuration)


DEFAULT_CONFIGURATION = Configurator.from_file(DEFAULTS_FOLDER / "quickclone.toml")
"""
An object storing the default configs for QuickClone.
"""


class SmartConfigurator(Configurator):
    """
    A child class of `Configurator` which can grab missing config items
    from `DEFAULT_CONFIGURATION`.
    """
    
    def __getitem__(self, key: t.Union[str, t.Iterable[str]]) -> t.Any:
        result = super().__getitem__(key)
        if result == "":
            return DEFAULT_CONFIGURATION[key]
        else:
            return result


def load_user_config(path: t.Optional[Path] = None) -> SmartConfigurator:
    """
    Load the user's config. If no config file is found, an empty
    `SmartConfigurator` will be returned and the defaults stored in
    `DEFAULT_CONFIGURATION` will be used instead.
    
    Parameters
    ----------
    path: Optional[Path] = None
        Path to the user's config files. The default argument should not be
        overridden unless necessary. If path is `None`, this function will
        try to look for the file at `USER_CONFIG_FILE`.
    
    Returns
    -------
    SmartConfigurator
        The user's configuration.
    """
    path = USER_CONFIG_FILE if path is None else path
    if path.exists() and path.is_file():
        return SmartConfigurator.from_file(path)
    else:
        return SmartConfigurator({})



from pathlib import Path
import typing as t


DEFAULTS_FOLDER: Path = Path(__file__).parent / "defaults"
"""
The path to the default configuration file.
"""


USER_CONFIG_FILE: Path = Path.home() / ".config" / "quickclone.toml"
"""
The path to the user's configuration file.
"""


USER_CACHE_FOLDER: Path = Path.home() / ".cache" / "quickclone"
"""
The path to the cache directory owned by the user to store QuickClone data.
"""

USER_HISTORY_CACHE_FILE: Path = USER_CACHE_FOLDER / "history.toml"
"""
The path to the cache file storing the user's usage history of QuickClone.
"""

CACHE_ITEMS: t.List[str] = ["history.toml"]
"""
List of file names in the cache folder.
"""



from __future__ import annotations
import typing as t

from .locators import BaseLocator, UrlAuthority


class ScpLocator(object):
    """
    An object for representing SCP locators. You will encounter such locators
    when you are trying to clone git repositories from github or gitlab using
    SSH. These locators come in this format:
    `<username>:<password>@<host>:<path>`.
    
    Parameters
    ----------
    host: str = ""
        The host where the remote repository is located.
        Examples: github.com, 1.1.1.1

    username: str = ""
        The username used to access the remote repository.
        Comes before the host and is separated from the latter by '@'.
        Examples: git (in git@github.com)

    password: str = ""
        Password of the user used to authenticate themselves.
        If present, must come after a username (separated by ':') and before
        a host name (separated by '@').
    
    path: str = ""
        The path to the remote repository
        (not referring to where the local clone is located).
        Comes after the host and separated from it by ':'.
        Examples: RenoirTan/QuickClone
        (in https://github.com/RenoirTan/QuickClone)
    
    **kwargs: Any
        Miscellaneous data.
    """
    
    def __init__(
        self,
        host: str = "",
        username: str = "",
        password: str = "",
        path: str = "",
        **kwargs
    ) -> None:
        self.host = host
        self.username = username
        self.password = password
        self.path = path
        self.kwargs = kwargs
    
    def __str__(self) -> str:
        authority_part = authority_part = str(
            UrlAuthority(self.get_host(), self.get_username(), self.get_password(), "")
        )
        path_part = f":{self.get_path()}" if self.get_path() != "" else ""
        return authority_part + path_part
        
    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"host={repr(self.host)}, "
            f"username={repr(self.username)}, "
            f"password={repr(self.password)}, "
            f"path={repr(self.path)}"
        )
    
    def __iter__(self) -> t.Generator[t.Tuple[str, str], None, None]:
        yield "host", self.get_host()
        yield "username", self.get_username()
        yield "password", self.get_password()
        yield "path", self.get_path()
    
    def get_host(self) -> str:
        """Get the host name of the website hosting the remote repository."""
        return self.host
    
    def get_username(self) -> str:
        """Get the username used to access the remote repository."""
        return self.username
    
    def get_password(self) -> str:
        """Get the password used to access the remote repository."""
        return self.password
    
    def get_path(self) -> str:
        """Get the path of the remote repository (in the website)."""
        return self.path
    
    @classmethod
    def from_locator(cls, locator: BaseLocator) -> ScpLocator:
        """
        Create an SCP locator from a normal locator (used in typical RFC 3986
        URLs).
        
        Parameters
        ----------
        cls: Type[ScpLocator]
            The `ScpLocator` class type.
        
        locator: BaseLocator
            The locator to be converted into an SCP locator.
        
        Returns
        -------
        ScpLocator
        """
        return cls(
            locator.get_host(),
            locator.get_username(),
            locator.get_password(),
            locator.get_path(),
            **locator.kwargs
        )



from __future__ import annotations
from pathlib import Path
import typing as t
import subprocess

import toml

from .common import USER_CACHE_FOLDER, USER_HISTORY_CACHE_FILE

AVAILABLE_CACHES: t.Set[str] = {"history"}
"""
The available cache data categories.
"""

_HISTORY_CACHE: t.Dict[str, t.Any] = dict()

def _create_file_if_not_exist(path: Path) -> None:
    if not path.exists():
        open(path, "x")

def load_caches(cache_names: t.Iterable[str]) -> int:
    """
    Load the data belonging to certain cache data categories.
    
    Parameters
    ----------
    cache_name: Iterable[str]
        List of cache categories to load up. Only the values that exist in
        `quickclone.config.cache.AVAILABLE_CACHES` are allowed.
    
    Raises
    ------
    ValueError
        If an invalid cache category name was given.
    
    Returns
    -------
    int
        The number of cache categories loaded.
    """
    subprocess.run(["mkdir", "-p", str(USER_CACHE_FOLDER)])
    count = 0
    for cache_name in cache_names:
        if cache_name == "history":
            _create_file_if_not_exist(USER_HISTORY_CACHE_FILE)
            global _HISTORY_CACHE
            _HISTORY_CACHE = toml.load(USER_HISTORY_CACHE_FILE)
        else:
            raise ValueError(f"Invalid cache name given: {cache_name}")
        count += 1
    return count

def _check_is_dict(obj: object) -> bool:
    return isinstance(obj, dict)

def _assert_is_dict(obj: object) -> None:
    assert _check_is_dict(obj), "Expected cache object to be an instance of a dictionary."

def _dump_cache(path: Path, cache: t.Dict[str, t.Any]) -> None:
    with open(path, "w") as f:
        toml.dump(cache, f)

def dump_caches(cache_names: t.Iterable[str]) -> int:
    """
    Dump the data belonging to certain cache data categories to their corresponding file.
    
    Parameters
    ----------
    cache_name: Iterable[str]
        List of cache categories to dump. Only the values that exist in
        `quickclone.config.cache.AVAILABLE_CACHES` are allowed.
    
    Raises
    ------
    ValueError
        If an invalid cache category name was given.
    
    Returns
    -------
    int
        The number of cache categories dumped.
    """
    subprocess.run(["mkdir", "-p", str(USER_CACHE_FOLDER)])
    count = 0
    for cache_name in cache_names:
        if cache_name == "history":
            _assert_is_dict(_HISTORY_CACHE)
            _dump_cache(USER_HISTORY_CACHE_FILE, _HISTORY_CACHE)
        else:
            raise ValueError(f"Invalid cache name given: {cache_name}")
    return count

def get_cache_value(desired: str) -> t.Optional[t.Any]:
    """
    Get cache value of desired key.
    """
    if desired == "last_clone":
        return _HISTORY_CACHE.get("last_clone")
    else:
        raise ValueError(f"Invalid desired={desired}")

def set_cache_value(desired: str, value: t.Optional[t.Any] = None) -> None:
    """
    Set cache value of desired key.
    """
    if desired == "last_clone":
        if isinstance(value, str) or isinstance(value, Path):
            _HISTORY_CACHE["last_clone"] = str(value)
        else:
            raise TypeError("Invalid type for last_clone")
    else:
        raise ValueError(f"Invalid desired={desired}")



from . import compatibility, config, delegation, remote

VERSION: str = "0.4.1"
"""
Current version of QuickClone
"""

NAME: str = "QuickClone"
"""
Name of the project/app.
"""

DESCRIPTION: str = "Command line utility for quickly cloning remote SCM repositories as succintly \
as possible."
"""
A description of the app.
"""

